/*	$Header: /usr/people/sam/fax/faxd/RCS/faxServerApp.h,v 1.38 93/04/04 13:44:25 sam Rel $
/*
 * Copyright (c) 1990, 1991, 1992, 1993 Sam Leffler
 * Copyright (c) 1991, 1992, 1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#ifndef _faxServerApp_
#define	_faxServerApp_
/*
 * Fax Queue Manager and Command Handler.
 */
#include <stdarg.h>
#include <Dispatch/iohandler.h>
#include "Str.h"

class FaxServer;
class FaxRequest;
class FaxMachineInfo;
class FaxRecvInfo;
class TIFF;

typedef unsigned int JobStatus;

/*
 * Jobs represent send requests in the queue.
 */
struct Job {
    enum {
	no_status,
	done,
	requeued,
	removed,
	timedout,
	no_formatter,
	format_failed,
	poll_rejected,
	poll_no_document,
	poll_failed,
	file_rejected,
    };
    Job*	next;		// linked list
    fxStr	file;		// queue file name
    time_t	tts;		// time to send job
    int		pri;		// priority

    Job(const fxStr& s, time_t t) : file(s) { tts = t; pri = 0; }
    Job(const Job& j) : file(j.file) { tts = j.tts; pri = 0; }
    Job(const Job* j) : file(j->file) { tts = j->tts; pri = 0; }
    ~Job() {}
};

/*
 * This class represents a thread of control that manages the
 * send queue, deals with system-related notification (sends
 * complete, facsimile received, errors), and delivers external
 * commands passed in through the command FIFOs.  This class is
 * also responsible for preparing documents for sending by
 * doing formatting tasks such as converting PostScript to TIFF.
 *
 * In a multi-threaded environment, this class represents a
 * separate thread of control.
 */
class faxServerApp : public IOHandler {
private:
    fxStr	appName;		// program name
    fxStr	device;			// modem device
    fxStr	devID;			// mangled device name
    fxStr	queueDir;		// spooling directory
    FaxServer*	server;			// active server app
    Job*	queue;			// job queue
    fxBool	running;		// server running
    int		requeueInterval;	// time between job retries
    u_long	currentTimeout;		// time associated with timer
    time_t	expectedTime;		// expected timer expiration
    time_t	jobStart;		// starting time for job
    time_t	fileStart;		// starting time for file
    int		fifo;			// fifo job queue interface
    int		devfifo;		// fifo device interface
    FaxRequest*	curreq;			// current job being processed

    static const fxStr fifoName;
    static const fxStr sendDir;
    static const fxStr recvDir;
    static const fxStr mailCmd;
    static const fxStr notifyCmd;
    static const fxStr ps2faxCmd;
    static const fxStr rtsCmd;
    static const fxStr transcriptCmd;

// miscellaneous stuff
    void	detachFromTTY();
    void	setRealIDs();
    void	logError(const char* fmt ...);
    void	vlogError(const char* fmt, va_list ap);
    void	traceServer(const char* fmt ...);
    void	traceQueue(const char* fmt ...);
    void	traceAtJobs(const char* fmt ...);
    const char*	fmtTime(time_t);
    void	record(const char* cmd,
		    const char* from, const char* to,
		    int rate, const char* protocol,
		    int npages, float time,
		    const char* reason);
    void	runCmd(const char* cmd, fxBool changeIDs = FALSE);

    static struct popenRecord* popenList;
    FILE*	popen(const char* cmd, const char* mode);
    int		pclose(FILE*);
// FIFO-related stuff
    int		openFIFO(const char* fifoName, int mode,
		    fxBool okToExist = FALSE);
    void	fifoMessage(const char* mesage);
// job management interfaces
    void	sendJob(FaxRequest* req);
    void	requestComplete(FaxRequest* req, fxBool notify);
    void	insertJob(Job* job);
    void	processJob(Job* job);
    void	deleteJob(const fxStr& name);
    Job*	removeJob(const fxStr& name);
    void	alterJob(const char* s);
// job preparation stuff
    void	scanQueueDirectory();
    void	scanQueue();
    void	startTimer(u_long sec);
    void	stopTimer();
    FaxRequest*	readQFile(const fxStr& filename, int fd);
    void	writeQFile(FaxRequest&);
    void	deleteRequest(JobStatus why, FaxRequest* req,
		    fxBool force = FALSE);
    void	notifySender(JobStatus why, FaxRequest& req);
    void	returnToSender(FILE*, FaxRequest&, fxBool includeTranscript);
    void	returnTranscript(FILE*, FaxRequest&);
    fxBool	copyFile(FILE* dst, FILE* src);
    JobStatus	convertPostScript(const fxStr& inFile, const fxStr& outFile,
		    FaxRequest& req, const FaxMachineInfo& info);
    JobStatus	checkFileFormat(const fxStr& file, const FaxMachineInfo& info,
		    fxStr& emsg);
    fxBool	checkPageFormat(TIFF*, const FaxMachineInfo& info, fxStr& emsg);
public:
    faxServerApp();
    ~faxServerApp();

    void	initialize(int argc, char** argv);
    void	open();
    void	close();

    fxBool	isRunning() const;
// Dispatcher hooks
    int		inputReady(int);
    void	timerExpired(long sec, long usec);
// notification interfaces used by FaxServer
    void	notifyModemReady();
    void	notifyDocumentSent(FaxRequest&, u_int fileIndex);
    void	notifyPollDone(FaxRequest&, u_int pollIndex, const char* file);
    void	notifySendDone(FaxRequest* req, u_int sigrate, const char* protoocol);
    void	notifyRecvDone(const FaxRecvInfo& req);
// system logging interfaces used by FaxServer
    void	vlogInfo(const char* fmt, va_list ap);
};
inline fxBool faxServerApp::isRunning() const { return running; }
#endif
