/*	$Header: /usr/people/sam/fax/sendfax/RCS/sendfax.c++,v 1.44 93/04/16 14:12:35 sam Rel $
/*
 * Copyright (c) 1990, 1991, 1992, 1993 Sam Leffler
 * Copyright (c) 1991, 1992, 1993 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Sam Leffler and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Sam Leffler and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 * 
 * IN NO EVENT SHALL SAM LEFFLER OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */
#include "FaxClient.h"
#include "FaxDB.h"
#include "StrArray.h"
#include "TypeRules.h"
#include "config.h"

#include "Dispatch/dispatcher.h"

#include <getopt.h>
#include <stdarg.h>
#include <ctype.h>
#include <paths.h>

#include <libc.h>
#include <osfcn.h>
#include <pwd.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
extern "C" {
#include <locale.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>		// XXX

    int close(int);
    int gethostname(char*, int);
    int access(const char*, int);
#ifdef __GNUC__
    char* strerror(int);
#endif
}

extern "C" int mkstemp(char*);

struct FileInfo : public fxObj {
    fxStr	name;
    const TypeRule* rule;
    fxBool	isTemp;
    fxBool	tagLine;

    FileInfo() { isTemp = FALSE; tagLine = TRUE; rule = NULL; }
    ~FileInfo() { if (isTemp) unlink((char*) name); }
};
fxDECLARE_ObjArray(FileInfoArray, FileInfo);

class sendFaxApp : public FaxClient {
private:
    fxStr	appName;		// for error messages
    fxStrArray	destNames;		// FAX destination names
    fxStrArray	destNumbers;		// FAX destination numbers
    fxStrArray	coverPages;		// cover pages
    TypeRules*	typeRules;		// file type/conversion database
    FileInfoArray files;		// files to send (possibly converted)
    fxStr	stdinTemp;		// temp file for collecting from pipe
    fxBool	pollCmd;		// do poll rather than send
    fxBool	coverSheet;		// if TRUE, prepend cover sheet
    fxBool	gotPermission;		// got response to checkPermission
    fxBool	permission;		// permission granted/denied
    u_int	verbose;
    fxStr	killtime;		// job's time to be killed
    fxStr	sendtime;		// job's time to be sent
    float	hres, vres;		// sending resolution (dpi)
    u_int	pageWidth;		// sending page width (mm)
    u_int	pageLength;		// sending page length (mm)
    int		totalPages;		// counted pages (for cover sheet)
    enum FaxNotify { no_notice, when_done, when_requeued } notify;
    FaxDB*	db;

    fxStr	mailBox;		// user@host return mail address
    fxStr	comments;		// comments on cover sheet
    fxStr	regarding;		// regarding line on cover sheet
    fxStr	from;			// command line from information
    fxStr	senderName;		// sender's full name
    fxStr	location;		// to's location
    fxStr	company;		// to's company

    static fxStr TypeRulesFile;
    static fxStr CoverSheetApp;
    static fxStr dbName;

    void	addDestination(const char* cp);
    fxStr	makeCoverPage(const fxStr& name, const fxStr& number);
    void	copyToTemporary(int fin, FileInfo& info);
    void	countTIFFPages(const char* name);
    void	estimatePostScriptPages(const char* name);
    const TypeRule* fileType(const char* filename);
    void	handleFile(const char* filename, FileInfo& info);
    void	setupSenderIdentity(const fxStr&);
    void	setupMailBox(const fxStr&);
    void	verifyPermission();
    fxStr	tildeExpand(const fxStr& filename);
    void	sendCoverDef(const char* name, const char* value);
    void	sendCoverLine(const char* va_alist ...);
    void	printError(const char* va_alist ...);
    void	usage();

    void	fatal(const char* fmt ...);
public:
    sendFaxApp();
    ~sendFaxApp();

    void	initialize(int argc, char** argv);
    void	open();

    void	recvConf(const char* cmd, const char* tag);
    void	recvEof();
    void	recvError(const int err);
};

fxStr sendFaxApp::TypeRulesFile(FAX_TYPERULES);
fxStr sendFaxApp::CoverSheetApp("faxcover");
fxStr sendFaxApp::dbName("~/.faxdb");

sendFaxApp::sendFaxApp()
{
    verbose = 0;
    pollCmd = FALSE;
    killtime = "now + 1 day";	// default time to kill the job
    coverSheet = TRUE;
    hres = 204;			// G3 standard
    vres = 98;			// default is low resolution
    pageWidth = 215;		// default is A4
    pageLength = 280;		// default is american reduced
    notify = no_notice;		// default is no email notification
    db = NULL;
    totalPages = 0;
    typeRules = NULL;
}

sendFaxApp::~sendFaxApp()
{
    if (stdinTemp.length())
	unlink((char*) stdinTemp);
    for (u_int i = 0; i < coverPages.length(); i++)
	unlink((char*) coverPages[i]);
    delete db;
    delete typeRules;
}

void
sendFaxApp::initialize(int argc, char** argv)
{
    extern int optind;
    extern char* optarg;
    int c;

    appName = argv[0];
    u_int l = appName.length();
    appName = appName.tokenR(l, '/');
    db = new FaxDB(tildeExpand(dbName));
    while ((c = getopt(argc, argv, "a:c:d:f:h:k:r:x:y:lmnpvABDWLR")) != -1)
	switch (c) {
	case 'A':			// ISO A4
	    pageWidth = 215;
	    pageLength = 297;
	    break;
	case 'B':			// ISO B4
	    pageWidth = 255;		// XXX don't know, just a guess
	    pageLength = 364;
	    break;
	case 'L':			// page length in mm
	    pageLength = atoi(optarg);
	    break;
	case 'W':			// page width in mm
	    pageWidth = atoi(optarg);
	    break;
	case 'c':			// cover sheet: comment field
	    comments = optarg;
	    break;
	case 'r':			// cover sheet: regarding field
	    regarding = optarg;
	    break;
	case 'd':			// destination name and number
	    addDestination(optarg);
	    break;
	case 'h':			// server's host
	    setHost(optarg);
	    break;
	case 'l':			// low resolution
	    vres = 98;
	    break;
	case 'm':			// medium resolution
	    vres = 196;
	    break;
	case 'n':			// no cover sheet
	    coverSheet = FALSE;
	    break;
	case 'p':			// submit polling request
	    pollCmd = TRUE;
	    break;
	case 'R':			// notify when requeued or done
	    notify = when_requeued;
	    break;
	case 'D':			// notify when done
	    notify = when_done;
	    break;
	case 'k':			// time to kill job
	    killtime = optarg;
	    break;
	case 'a':			// time at which to transmit job
	    sendtime = optarg;
	    break;
	case 'v':
	    verbose++;
	    break;
	case 'f':			// sender's identity
	    from = optarg;
	    break;
	case 'x':			// cover page: to's company
	    company = optarg;
	    break;
	case 'y':			// cover page: to's location
	    location = optarg;
	    break;
	case '?':
	    usage();
	    /*NOTREACHED*/
	}
    if (destNumbers.length() == 0) {
	printError("No destination specified");
	usage();
    }
    if (verbose) {
	setvbuf(stdout, NULL, _IOLBF, BUFSIZ);
	if (verbose > 1)
	    setVerbose(TRUE);
    }
    typeRules = TypeRules::read(fxStr(FAX_LIBDIR) | "/" | TypeRulesFile);
    if (!typeRules)
	fatal("Unable to setup file typing and conversion rules");
    typeRules->setVerbose(verbose > 0);
    setupSenderIdentity(from);
    if (optind < argc) {
	files.resize(argc-optind);
	for (u_int i = 0; optind < argc; i++, optind++)
	    handleFile(argv[optind], files[i]);
    } else if (!pollCmd) {
	files.resize(1);
	copyToTemporary(fileno(stdin), files[0]);
    }
    // suppress cover page if we're just doing a poll
    if (pollCmd && files.length() == 0)
	coverSheet = FALSE;
    if (coverSheet) {
	coverPages.resize(destNumbers.length());
	for (u_int i = 0; i < destNumbers.length(); i++)
	    coverPages[i] = makeCoverPage(destNames[i], destNumbers[i]);
    }
}

void
sendFaxApp::usage()
{
    fxFatal("usage: %s"
	" [-d destination]"
	" [-h host[:modem]]"
	" [-k kill-time]\n"
	"    "
	" [-a time-to-send]"
	" [-c comments]"
	" [-r regarding]\n"
	"    "
	" [-W pagewidth]"
	" [-L pagelength]"
	" [-lmnpvABDR]"
	" [files]",
	(char*) appName);
}

/*
 * Invoke the cover page generation program.
 */
fxStr
sendFaxApp::makeCoverPage(const fxStr& name, const fxStr& number)
{
    fxStr templ(_PATH_TMP "sndfaxXXXXXX");
    int fd = mkstemp(templ);
    if (fd < 0)
	fatal("%s: Can not create temporary file for cover sheet",
	    (char*) templ);
    fxStr cmd(CoverSheetApp);
    cmd.append(" -n \"" | number | "\"");
    cmd.append(" -t \"" | name | "\"");
    cmd.append(" -f \"" | senderName | "\"");
    if (comments != "")
	cmd.append(" -c \"" | comments | "\"");
    if (regarding != "")
	cmd.append(" -r \"" | regarding | "\"");
    if (location != "")
	cmd.append(" -l \"" | location | "\"");
    if (company != "")
	cmd.append(" -x \"" | company | "\"");
    if (totalPages > 0)
	cmd.append(" -p " | fxStr(totalPages, "%u"));
    if (verbose)
	printf("COVER SHEET \"%s\"\n", (char*) cmd);
    FILE* fp = popen(cmd, "r");
    if (fp != NULL) {
	// copy prototype cover page
	char line[1024];
	while (fgets(line, sizeof (line)-1, fp))
	    (void) write(fd, line, strlen(line));
	if (pclose(fp) == 0) {
	    ::close(fd);
	    return (templ);
	}
    }
    unlink((char*) templ);
    fatal("Error creating cover sheet; command was \"%s\"", (char*) cmd);
    /*NOTREACHED*/
}

/*
 * Add a new destination name and number.
 */
void
sendFaxApp::addDestination(const char* cp)
{
    fxStr recipient;
    const char* tp = strchr(cp, '@');
    if (tp) {
	recipient = fxStr(cp, tp-cp);
	cp = tp+1;
    } else
	recipient = "";
    fxStr dest(cp);
    if (db && dest.length() > 0) {
	fxStr name;
	FaxDBRecord* r = db->find(dest, &name);
	if (r) {
	    if (recipient == "")
		recipient = name;
	    dest = r->find(FaxDB::numberKey);
	}
    }
    if (dest.length() == 0) {
	printError("Null destination for \"%s\"", cp);
	exit(-1);
    }
    destNames.append(recipient);
    destNumbers.append(dest);
}

/*
 * Expand a filename that might have `~' in it.
 */
fxStr
sendFaxApp::tildeExpand(const fxStr& filename)
{
    fxStr path(filename);
    if (filename.length() > 1 && filename[0] == '~') {
	path.remove(0);
	char* cp = getenv("HOME");
	if (!cp || *cp == '\0') {
	    struct passwd* pwd = getpwuid(getuid());
	    if (!pwd)
		fatal("Can not figure out who you are");
	    cp = pwd->pw_dir;
	}
	path.insert(cp);
    }
    return (path);
}

/*
 * Check parsed sender identity against identity obtained
 * by looking at the real uid.  If they're different and
 * we are not running as a "trusted user", abort.
 */
void
sendFaxApp::verifyPermission()
{
    uid_t uid = getuid();
    if (uid == 0)
	return;
    passwd* pwd = getpwuid(uid);
    fxStr name(pwd ? pwd->pw_name : "nobody");
    if (name == "daemon" || name == "uucp")
	return;
    // check specified address against real uid's address
    fxStr user(mailBox.head(mailBox.next(0, '@')));
    user.remove(0, user.nextR(user.length(), '!'));
    if (user != getUserName())
	fatal("Unauthorized attempt to set sender's identity (from %s, uid %s)",
	    (char*) user, (char*) getUserName());
}

/*
 * Create the mail address for a local user.
 */
void
sendFaxApp::setupMailBox(const fxStr& user)
{
    char hostname[64];
    (void) gethostname(hostname, sizeof (hostname));
    struct hostent* hp = gethostbyname(hostname);
    mailBox = user | "@" | (hp ? hp->h_name : hostname);
}

/*
 * Setup the sender's identity.
 */
void
sendFaxApp::setupSenderIdentity(const fxStr& from)
{
    FaxClient::setupUserIdentity();		// client identity

    if (from != "") {
	u_int l = from.next(0, '<');
	if (l == from.length()) {
	    l = from.next(0, '(');
	    if (l != from.length()) {		// joe@foobar (Joe Schmo)
		mailBox = from.head(l);
		l++; senderName = from.token(l, ')');
	    } else {				// joe
		setupMailBox(from);
		if (from == getUserName())
		    senderName = getSenderName();
		else
		    senderName = "";
	    }
	} else {				// Joe Schmo <joe@foobar>
	    senderName = from.head(l);
	    l++; mailBox = from.token(l, '>');
	}
	if (senderName == "" && mailBox != "") {
	    /*
	     * Mail address, but no "real name"; construct one from
	     * the account name.  Do this by first stripping anything
	     * to the right of an '@' and then stripping any leading
	     * uucp patch (host!host!...!user).
	     */
	    senderName = mailBox;
	    senderName.resize(senderName.next(0, '@'));
	    senderName.remove(0, senderName.nextR(senderName.length(), '!'));
	}

	// strip and leading&trailing white space
	senderName.remove(0, senderName.skip(0, " \t"));
	senderName.resize(senderName.skipR(senderName.length(), " \t"));
	mailBox.remove(0, mailBox.skip(0, " \t"));
	mailBox.resize(mailBox.skipR(mailBox.length(), " \t"));
	if (senderName == "" || mailBox == "")
	    fatal("Malformed (null) sender name or mail address");

	verifyPermission();
    } else {
	senderName = getSenderName();
	setupMailBox(getUserName());
    }
}

/*
 * Process a file submitted for transmission.
 */
void
sendFaxApp::handleFile(const char* filename, FileInfo& info)
{
    if (strcmp(filename, "-") == 0) {
	// treat - as "read from stdin"
	copyToTemporary(fileno(stdin), info);
	return;
    }
    info.rule = fileType(filename);
    if (info.rule->getCmd() != "") {	// conversion required
	fxStr temp(_PATH_TMP "sndfaxXXXXXX");
	mktemp((char*) temp);
	fxStr sysCmd = info.rule->getFmtdCmd(filename, temp,
		hres, vres, "1", pageWidth, pageLength);
	if (verbose)
	    printf("CONVERT \"%s\"\n", (char*) sysCmd);
	if (system((char*) sysCmd) != 0) {
	    unlink((char*) temp);
	    fatal("Error converting data; command was \"%s\"", (char*)sysCmd);
	}
	info.name = temp;
	info.isTemp = TRUE;
    } else {				// already postscript or tiff
	info.name = filename;
	info.isTemp = FALSE;
    }
    switch (info.rule->getResult()) {
    case TypeRule::TIFF:
	countTIFFPages(info.name);
	break;
    case TypeRule::POSTSCRIPT:
	estimatePostScriptPages(info.name);
	break;
    }
}

/*
 * Return a TypeRule for the specified file.
 */
const TypeRule*
sendFaxApp::fileType(const char* filename)
{
    struct stat sb;
    int fd = ::open(filename, O_RDONLY);
    if (fd < 0)
	fatal("%s: Can not open file", filename);
    if (fstat(fd, &sb) < 0)
	fatal("%s: Can not stat file", filename);
    if ((sb.st_mode & S_IFMT) != S_IFREG)
	fatal("%s: Not a regular file", filename);
    char buf[512];
    int cc = read(fd, buf, sizeof (buf));
    if (cc == 0)
	fatal("%s: Empty file", filename);
    const TypeRule* tr = typeRules->match(buf, cc);
    if (!tr)
	fatal("%s: Can not determine file type", filename);
    if (tr->getResult() == TypeRule::ERROR) {
	fxStr emsg(tr->getErrMsg());
	fatal("%s: %s", filename, (char*) emsg);
    }
    return tr;   
}

/*
 * Copy data from fin to a temporary file.
 */
void
sendFaxApp::copyToTemporary(int fin, FileInfo& info)
{
    fxStr templ(_PATH_TMP "sndfaxXXXXXX");
    int fd = mkstemp(templ);
    if (fd < 0)
	fatal("%s: Can not create temporary file", (char*) templ);
    int cc, total = 0;
    char buf[16*1024];
    while ((cc = read(fin, buf, sizeof (buf))) > 0) {
	if (write(fd, buf, cc) != cc) {
	    unlink((char*) templ);
	    fatal("%s: write error", (char*) templ);
	}
	total += cc;
    }
    ::close(fd);
    if (total == 0) {
	unlink((char*) templ);
	fatal("No input data; tranmission aborted");
    }
    handleFile(templ, info);
    /*
     * If the temp file contained data that was converted, then
     * we can unlink it now.  Otherwise, it'll be unlinked when
     * we exit as a consequence of our marking it ``temporary''.
     */
    if (info.name != templ)
	unlink((char*) templ);
    else
	info.isTemp = TRUE;
}

#include "tiffio.h"

/*
 * Count the number of ``pages'' in a TIFF file.
 */
void
sendFaxApp::countTIFFPages(const char* filename)
{
    TIFF* tif = TIFFOpen(filename, "r");
    if (tif) {
	do {
	    totalPages++;
	} while (TIFFReadDirectory(tif));
	TIFFClose(tif);
    }
}

/*
 * Count the number of pages in a PostScript file.
 * We can really only estimate the number as we
 * depend on the DSC comments to figure this out.
 */
void
sendFaxApp::estimatePostScriptPages(const char* filename)
{
    FILE* fd = fopen(filename, "r");
    if (fd != NULL) {
	char line[2048];
	if (fgets(line, sizeof (line)-1, fd) != NULL) {
	    /*
	     * We only consider ``conforming'' PostScript documents.
	     */
	    if (line[0] == '%' && line[1] == '!') {
		int npagecom = 0;	// # %%Page comments
		int npages = 0;		// # pages according to %%Pages comments
		while (fgets(line, sizeof (line)-1, fd) != NULL) {
		    int n;
		    if (strncmp(line, "%%Page:", 7) == 0)
			npagecom++;
		    else if (sscanf(line, "%%%%Pages: %u", &n) == 1)
			npages += n;
		}
		/*
		 * Believe %%Pages comments over counting of %%Page comments.
		 */
		if (npages > 0)
		    totalPages += npages;
		else if (npagecom > 0)
		    totalPages += npagecom;
	    }
	}
	fclose(fd);
    }
}

void
sendFaxApp::open()
{
    if (callServer() == FaxClient::Failure)
	fatal("Could not call server");
    startRunning();
    /*
     * Explicitly check for permission to submit a job
     * before sending the input documents.  This way we
     * we avoid sending a load of stuff just to find out
     * that the user/host is not permitted to submit jobs.
     */
    sendLine("checkPerm", "send");
    permission = gotPermission = FALSE;
    while (isRunning() && !getPeerDied() && !gotPermission)
	Dispatcher::instance().dispatch();
    if (!permission)
	return;

    /*
     * Transfer the document files first so that they
     * can be referenced multiple times for different
     * destinations.
     */
    for (u_int i = 0; i < files.length(); i++) {
	const FileInfo& info = files[i];
	sendData(info.rule->getResult() == TypeRule::TIFF ?
	    "tiff" : "postscript", info.name);
    }
    if (pollCmd)
	sendLine("poll", "");
    for (i = 0; i < destNumbers.length(); i++) {
	sendLine("begin", i);
	if (sendtime.length() > 0)
	    sendLine("sendAt", sendtime);
	sendLine("killtime", killtime);
	sendLine("number", destNumbers[i]);
	sendLine("sender", senderName);
	sendLine("mailaddr", mailBox);
	sendLine("resolution", (int) vres);
	sendLine("pagewidth", pageWidth);
	sendLine("pagelength", pageLength);
	if (notify == when_done)
	    sendLine("notify", "when done");
	else if (notify == when_requeued)
	    sendLine("notify", "when requeued");
	if (coverSheet) {
	    FILE* fp = fopen(coverPages[i], "r");
	    if (fp != NULL) {
		sendLine("cover", 1);	// cover sheet sub-protocol version 1
		// copy prototype cover page
		char line[1024];
		while (fgets(line, sizeof (line)-1, fp)) {
		    char* cp = strchr(line, '\n');
		    if (cp)
			*cp = '\0';
		    sendCoverLine("%s", line);
		}
		sendLine("..\n");
		fclose(fp);
	    } else
		printError("Beware, the cover page for \"%s\" disappeared; "
		    "none was sent", (char*) destNumbers[i]);
	}
	sendLine("end", i);
    }
    sendLine(".\n");
}

void
sendFaxApp::sendCoverLine(const char* va_alist ...)
#define	fmt va_alist
{
    va_list ap;
    va_start(ap, va_alist);
    char buf[4096];
    vsprintf(buf, fmt, ap);
    va_end(ap);
    sendLine("!", buf);
}
#undef fmt

void
sendFaxApp::printError(const char* va_alist ...)
#define	fmt va_alist
{
    va_list ap;
    va_start(ap, va_alist);
    fprintf(stderr, "%s: ", (char*) appName);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    fprintf(stderr, ".\n");
}
#undef fmt

#define	isCmd(s)	(strcasecmp(s, cmd) == 0)

void
sendFaxApp::recvConf(const char* cmd, const char* tag)
{
    if (isCmd("job")) {
	int len = files.length();
	printf("request id is %s for host %s (%d %s)\n",
	    tag, (char*) getHost(), len, len > 1 ? "files" : "file");
    } else if (isCmd("permission")) {
	gotPermission = TRUE;
	permission = (strcasecmp(tag, "granted") == 0);
    } else if (isCmd("error")) {
	printf("%s\n", tag);
    } else
	printf("Unknown status message \"%s:%s\"\n", cmd, tag);
}

void
sendFaxApp::recvEof()
{
    stopRunning();
}

void
sendFaxApp::recvError(const int err)
{
    printError("Fatal socket read error: %s", strerror(err));
    stopRunning();
}

fxIMPLEMENT_ObjArray(FileInfoArray, FileInfo);

static sendFaxApp* app;

int
main(int argc, char** argv)
{
#ifdef LC_CTYPE
    setlocale(LC_CTYPE, "");
#endif
    app = new sendFaxApp;
    app->initialize(argc, argv);
    app->open();
    while (app->isRunning())
	Dispatcher::instance().dispatch();
    delete app;
    return 0;
}

static void
vfatal(FILE* fd, const char* fmt, va_list ap)
{
    vfprintf(fd, fmt, ap);
    va_end(ap);
    fputs(".\n", fd);
    delete app;				// cleanup temporary files
    exit(-1);
}

void
fxFatal(const char* va_alist ...)
#define	fmt va_alist
{
    va_list ap;
    va_start(ap, fmt);
    vfatal(stderr, fmt, ap);
    /*NOTTEACHED*/
}
#undef fmt

void
sendFaxApp::fatal(const char* va_alist ...)
#define	fmt va_alist
{
    fprintf(stderr, "%s: ", (char*) appName);
    va_list ap;
    va_start(ap, fmt);
    vfatal(stderr, fmt, ap);
    /*NOTTEACHED*/
}
#undef fmt
